/* (C)opyright MMVI-MMVII Anselm R. Garbe <garbeam at gmail dot com>
 * (C)opyright MMVI-MMVII Sander van Dijk <a dot h dot vandijk at gmail dot com>
 * (C)opyright MMVII      Christian Dietrich <stettberger at gmx dot com>
 * See LICENSE file for license details.
 */

#include "dinput.h"
#include <ctype.h>
#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define CLEANMASK(mask) (mask & ~(numlockmask | LockMask))
#define alptr (lines[aline]->l)
#define apptr (lines[aline]->l + lines[aline]->p)

typedef struct Line Line;
struct Line {
  char *l; /* data */
  unsigned int p; /* cursor pointer */
};
typedef struct InputWindow InputWindow;
struct InputWindow {
  Window win;
  unsigned int mh;
  unsigned int mw;
  unsigned int lines;
  unsigned int lh;
};
/* static */
static char *prompt = NULL;
static Line **lines = NULL;
static unsigned int aline = 0;
static unsigned int maxline = 0;
static int ret = 0;
static unsigned int promptw = 0;
static unsigned int numlockmask = 0;
static Bool running = True;
static Bool rstdin = True;
static Bool bottom = False;
static Window root;
static InputWindow *input;

void
drawmenu(void) {
        char cursor[] = {0,0};
        char *p = alptr;
        int i, cy = 0, cx = promptw;
        
        /* calculate lines */
        input->lines = (textnw(alptr, strlen(alptr)) + promptw)
              / (input->mw - dc.font.height - dc.font.ascent) + 1;
        if (input->lines < 1) input->lines = 1;

        /* Window size has changed */
        if (input->lines * input->lh != input->mh) {
          input->mh = input->lines * input->lh;
          XResizeWindow(dpy, input->win, input->mw, input->mh);
	  XFreePixmap(dpy, dc.drawable);
	  dc.drawable = XCreatePixmap(dpy, root, input->mw, input->mh, DefaultDepth(dpy, screen));
          if (bottom) 
            XMoveWindow(dpy, input->win, 0, DisplayHeight(dpy, screen) - input->mh);
        }
	
        dc.x = 0;
	dc.y = 0;
	dc.w = input->mw;
	dc.h = input->lh;
        for (i = 0; i< input->lines; i++) {
          drawtext(NULL, dc.norm, 0);
          dc.y += input->lh;
        }
        dc.y = 0;

	/* print prompt? */
	if(promptw) {
		dc.w = promptw;
		drawtext(prompt, dc.sel, strlen(prompt));
	}
	dc.x += promptw;
        dc.w = input->mw - promptw;

        /* Draw text */
        do {
          i = width2len(p, dc.w); 
          if ( apptr >= p && apptr <= p + i + 1) {
            cy = dc.y;
            cx = dc.x + textnw(p, apptr - p) + dc.font.height/2;
          }
          drawtext(p, dc.norm, i);
          p += i;
          dc.w = input->mw;
          dc.x = 0;
          dc.y += dc.h;
        } while (*p);
        
        /* Draw cursor */
        dc.y = cy;
        dc.x = cx;
        cursor[0] = apptr[0]?apptr[0]:' ';
        dc.w = textnw(cursor, 1) ;
        draw_string(cursor, dc.sel);
        XCopyArea(dpy, dc.drawable, input->win, dc.gc, 0, 0, input->mw, input->mh, 0, 0);
	XFlush(dpy);
}

static Bool
grabkeyboard(void) {
  unsigned int len;

  for(len = 1000; len; len--) {
    if(XGrabKeyboard(dpy, root, True, GrabModeAsync, GrabModeAsync, CurrentTime)
        == GrabSuccess)
      break;
    usleep(1000);
  }
  return len > 0;
}

static unsigned long
initcolor(const char *colstr) {
	Colormap cmap = DefaultColormap(dpy, screen);
	XColor color;

	if(!XAllocNamedColor(dpy, cmap, colstr, &color, &color))
		eprint("error, cannot allocate color '%s'\n", colstr);
	return color.pixel;
}

static void
initfont(const char *fontstr) {
	char *def, **missing;
	int i, n;

	missing = NULL;
	if(dc.font.set)
		XFreeFontSet(dpy, dc.font.set);
	dc.font.set = XCreateFontSet(dpy, fontstr, &missing, &n, &def);
	if(missing)
		XFreeStringList(missing);
	if(dc.font.set) {
		XFontSetExtents *font_extents;
		XFontStruct **xfonts;
		char **font_names;
		dc.font.ascent = dc.font.descent = 0;
		font_extents = XExtentsOfFontSet(dc.font.set);
		n = XFontsOfFontSet(dc.font.set, &xfonts, &font_names);
		for(i = 0, dc.font.ascent = 0, dc.font.descent = 0; i < n; i++) {
			if(dc.font.ascent < (*xfonts)->ascent)
				dc.font.ascent = (*xfonts)->ascent;
			if(dc.font.descent < (*xfonts)->descent)
				dc.font.descent = (*xfonts)->descent;
			xfonts++;
		}
	}
	else {
		if(dc.font.xfont)
			XFreeFont(dpy, dc.font.xfont);
		dc.font.xfont = NULL;
		if(!(dc.font.xfont = XLoadQueryFont(dpy, fontstr)))
			eprint("error, cannot load font: '%s'\n", fontstr);
		dc.font.ascent = dc.font.xfont->ascent;
		dc.font.descent = dc.font.xfont->descent;
	}
	dc.font.height = dc.font.ascent + dc.font.descent;
}


static void
kpress(XKeyEvent * e) {
	char buf[32];
	int num;
	unsigned int len;
	KeySym ksym;

	len = strlen(alptr);
	buf[0] = 0;
	num = XLookupString(e, buf, sizeof buf, &ksym, 0);
	if(IsFunctionKey(ksym) || IsKeypadKey(ksym)
			|| IsMiscFunctionKey(ksym) || IsPFKey(ksym)
			|| IsPrivateKeypadKey(ksym))
		return;
	/* first check if a control mask is omitted */
	if(e->state & ControlMask) {
		switch (ksym) {
		default:	/* ignore other control sequences  */
			return;
		case XK_bracketleft:
			ksym = XK_Escape;
			break;
		case XK_h:
		case XK_H:
			ksym = XK_BackSpace;
			break;
		case XK_j:
		case XK_J:
			ksym = XK_Return;
			break;
                case XK_u:
                case XK_U:
                        alptr[0] = 0;
                        lines[aline]->p = 0;
                        drawmenu();
                        return;
		}
	}
	if(CLEANMASK(e->state) & Mod1Mask) {
		switch(ksym) {
		default: return;
		case XK_h:
			ksym = XK_Left;
			break;
		case XK_l:
			ksym = XK_Right;
			break;
		case XK_j:
			ksym = XK_Next;
			break;
		case XK_k:
			ksym = XK_Prior;
			break;
		case XK_w:
			ksym = XK_Delete;
			break;
		case XK_g:
			ksym = XK_Home;
			break;
		case XK_G:
			ksym = XK_End;
			break;
		}
	}
	switch(ksym) {
	default:
                if(num && !iscntrl((int) buf[0])) {
		  buf[num] = 0;
                  int len = strlen(alptr);
                  alptr = erealloc(alptr, len + num + 1);
                  memmove(apptr + num, apptr, strlen(apptr) + 1);
                  memcpy(apptr, buf, num);
                  lines[aline]->p += num;
		}
		break;
        case XK_Up:
                if (aline > 0)
                  aline--;
                break;
        case XK_Down:
                if ( aline < maxline )
                  aline++;
                break;
        case XK_Delete:
                memmove(apptr, apptr + 1, strlen(apptr));
                break;
	case XK_BackSpace:
                if (lines[aline]->p) {
                  memmove(apptr - 1, apptr, strlen(apptr) + 1);
                  lines[aline]->p--;
                }
		break;
	case XK_End:
                lines[aline]->p = strlen(alptr);
		break;
	case XK_Escape:
		ret = 1;
		running = False;
		break;
	case XK_Home:
                lines[aline]->p = 0;
		break;
	case XK_Next:
	case XK_Right:
                lines[aline]->p += apptr[0]?1:0;
		break;
	case XK_Return:
		fprintf(stdout, "%s", alptr);
		fflush(stdout);
		running = False;
		break;
	case XK_Prior:
	case XK_Left:
                if (lines[aline]->p > 0)
                  lines[aline]->p--;
		break;
	}
	drawmenu();
}

void 
readstdin(void) 
{
  char *line;
  int len = 0;
  lines = emalloc(sizeof(Line *) * 2);
  while (rstdin && (line = rline())) {
    if (! line[0]) {
      free(line);
      continue;
    }
    lines[len] = emalloc(sizeof(Line));
    lines[len]->l = line;
    len++;
    lines = erealloc(lines, (len + 2) * sizeof(Line *));
    lines[len] = NULL;
  }
  lines = erealloc(lines, ((len + 2) * sizeof(Line *)));
  lines[len + 1] = NULL;
  lines[len] = emalloc(sizeof(Line));
  lines[len]->l = estrdup("");
  maxline = aline = len;
}

static void
usage(void) {
	eprint("usage: dinput [-b] [-fn <font>] [-nb <color>] [-nf <color>] [-ns]\n"
		"             [-p <prompt>] [-sb <color>] [-sf <color>] [-v]\n");
}

/* extern */

int screen;
Display *dpy;
DC dc = {0};

int
main(int argc, char *argv[]) {
	char *font = FONT;
	char *normbg = NORMBGCOLOR;
	char *normfg = NORMFGCOLOR;
	char *selbg = SELBGCOLOR;
	char *selfg = SELFGCOLOR;
	int i, j;
	XEvent ev;
	XModifierKeymap *modmap;
	XSetWindowAttributes wa;
        input = emalloc(sizeof(InputWindow));

	/* command line args */
	for(i = 1; i < argc; i++)
		if(!strncmp(argv[i], "-b", 3)) {
			bottom = True;
		}
                else if(!strncmp(argv[i], "-ns", 3)) {
			rstdin = False;
		}
		else if(!strncmp(argv[i], "-fn", 4)) {
			if(++i < argc) font = argv[i];
		}
                
		else if(!strncmp(argv[i], "-nb", 4)) {
			if(++i < argc) normbg = argv[i];
		}
		else if(!strncmp(argv[i], "-nf", 4)) {
			if(++i < argc) normfg = argv[i];
		}
		else if(!strncmp(argv[i], "-p", 3)) {
			if(++i < argc) prompt = argv[i];
		}
		else if(!strncmp(argv[i], "-sb", 4)) {
			if(++i < argc) selbg = argv[i];
		}
		else if(!strncmp(argv[i], "-sf", 4)) {
			if(++i < argc) selfg = argv[i];
		}
		else if(!strncmp(argv[i], "-v", 3))
			eprint("dinput-"VERSION", (C)opyright MMVI-MMVII Anselm R. Garbe\n");
		else
			usage();
	setlocale(LC_CTYPE, "");
	dpy = XOpenDisplay(0);
	if(!dpy)
		eprint("dinput: cannot open display\n");
	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);
	/* style */
	dc.norm[ColBG] = initcolor(normbg);
	dc.norm[ColFG] = initcolor(normfg);
	dc.sel[ColBG] = initcolor(selbg);
	dc.sel[ColFG] = initcolor(selfg);
	initfont(font);
	/* menu window */
	wa.override_redirect = 1;
	wa.background_pixmap = ParentRelative;
	wa.event_mask = ExposureMask | ButtonPressMask | KeyPressMask;
	input->mw = DisplayWidth(dpy, screen);
	input->lh = input->mh = dc.font.height + 2;
	input->win = XCreateWindow(dpy, root, 0,
			bottom ? DisplayHeight(dpy, screen) - input->mh : 0, input->mw, 
                        input->mh, 0, DefaultDepth(dpy, screen), CopyFromParent,
			DefaultVisual(dpy, screen),
			CWOverrideRedirect | CWBackPixmap | CWEventMask, &wa);
	/* pixmap */
	dc.drawable = XCreatePixmap(dpy, root, input->mw, input->mh, DefaultDepth(dpy, screen));
	dc.gc = XCreateGC(dpy, root, 0, 0);
	XSetLineAttributes(dpy, dc.gc, 1, LineSolid, CapButt, JoinMiter);
	if(!dc.font.set)
		XSetFont(dpy, dc.gc, dc.font.xfont->fid);
	XMapRaised(dpy, input->win);
	if(isatty(STDIN_FILENO)) {
		readstdin();
		running = grabkeyboard();
	}
	else { /* prevent keypress loss */
		running = grabkeyboard();
		readstdin();
	}
	drawmenu();
	/* init modifier map */
	modmap = XGetModifierMapping(dpy);
	for(i = 0; i < 8; i++)
		for(j = 0; j < modmap->max_keypermod; j++) {
			if(modmap->modifiermap[i * modmap->max_keypermod + j]
			== XKeysymToKeycode(dpy, XK_Num_Lock))
				numlockmask = (1 << i);
		}
	XFreeModifiermap(modmap);
	if(prompt)
		promptw = textw(prompt);
	if(promptw > input->mw / 5)
		promptw = input->mw / 5;
	XSync(dpy, False);

	/* main event loop */
	while(running && !XNextEvent(dpy, &ev))
		switch (ev.type) {
		default:	/* ignore all crap */
			break;
		case KeyPress:
			kpress(&ev.xkey);
			break;
		case Expose:
			if(ev.xexpose.count == 0)
				drawmenu();
			break;
		}

	/* cleanup */
	for (i = 0; lines[i]; i++) {
          free(lines[i]->l);
          free(lines[i]);
        }
        free(lines);
	
        if(dc.font.set)
		XFreeFontSet(dpy, dc.font.set);
	else
		XFreeFont(dpy, dc.font.xfont);
	XFreePixmap(dpy, dc.drawable);
	XFreeGC(dpy, dc.gc);
	XDestroyWindow(dpy, input->win);
	XUngrabKeyboard(dpy, CurrentTime);
	XCloseDisplay(dpy);
        free(input);

	return ret;
}
