/* (C)opyright MMVI-MMVII Anselm R. Garbe <garbeam at gmail dot com>
 * (C)opyright MMVII      Christian Dietrich <stettberger at gmx dot com>
 * See LICENSE file for license details.
 */
#include "dinput.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void *
emalloc(unsigned int size) {
	void *res = calloc(1, size);

	if(!res)
		eprint("fatal: could not malloc() %u bytes\n", size);
	return res;
}

void
eprint(const char *errstr, ...) {
	va_list ap;

	va_start(ap, errstr);
	vfprintf(stderr, errstr, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

char *
estrdup(const char *str) {
	void *res = strdup(str);

	if(!res)
		eprint("fatal: could not malloc() %u bytes\n", strlen(str));
	return res;
}

char *
rline(void)
{
  char *str = emalloc(1024);
  char buf[1024];
  int len = 1024;
  do {
    if (!fgets(buf, sizeof buf, stdin)) {
      if (*str) {
        str[strlen(str)-1] = 0;
        return str;
      }
      else {
        free(str);
        return NULL;
      }
    }
    if (len < strlen(str) + strlen(buf))
      str = erealloc(str, strlen(str) + strlen(buf) + 1);
    strcat(str, buf);
  } while(str[strlen(str)-1] != '\n');
  str[strlen(str)-1] = 0;
  return str;
}

void *
erealloc(void *ptr, size_t size)
{
  void *res = realloc(ptr, size);
  if(!res)
    eprint("fatal: could not realloc %u bytes\n", size);
  return res;
}
